/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "common.h"

#include <cstdio>
#include <cstring>

#include <algorithm>
#include <vector>

std::vector<const char*> get_cmd_args(char** begin, char** end)
{
  std::vector<const char*> list;
  for (char** itr = begin+1; itr != end; itr++)
  {
    if (strncmp(*itr, "--", 2) != 0)
      list.push_back(*itr);
  }
  return list;
}

bool cmd_option_exists(char** begin, char** end, const std::string& option)
{
  return std::find(begin, end, option) != end;
}

static void help(const char *argv0)
{
  const char* helpdesc = 
    "Usage: %s [args] <root paths or pkg name>\n"
    "      --help       - Display this screen\n"
    "      --system     - Create NI under System DLLs\n"
    "      --dll        - Create NI for DLL\n"
    "      --pkg        - Create NI for package\n"
    "\n"
    "Example:\n"
    "Create native image for dlls and exes under platform directories\n"
    "%s --system\n"
    "Create native image for dll\n"
    "%s --dll /usr/bin/Tizen.Runtime.Coreclr.dll\n"
    "Create native image under the package's bin and lib directory\n"
    "%s --pkg org.tizen.FormsGallery\n\n";
  printf(helpdesc, argv0, argv0, argv0, argv0);
}

int main(int argc, char* argv[])
{
  bool pkg_mode = false;
  bool dll_mode = false;

  if (cmd_option_exists(argv, argv+argc, "--help"))
  {
    help(argv[0]);
    return 0;
  }
  else if (cmd_option_exists(argv, argv+argc, "--system"))
  {
    create_ni_platform();
    return 0;
  }
  else if (cmd_option_exists(argv, argv+argc, "--dll"))
  {
    dll_mode = true;
  }
  else if (cmd_option_exists(argv, argv+argc, "--pkg"))
  {
    pkg_mode = true;
  }
  else {
    help(argv[0]);
    return 1;
  }

  std::vector<const char*> args = get_cmd_args(argv, argv+argc);

  if (args.size() < 1)
  {
    if (pkg_mode)
      fprintf(stderr, "Package name is missed\n");
    else if (dll_mode)
      fprintf(stderr, "DLL path is missed\n");
    help(argv[0]);
    return 1;
  }

  if (pkg_mode)
  {
    for (const char* pkg : args)
    {
      if (create_ni_under_pkg_root(pkg) != 0)
      {
        fprintf(stderr, "Failed to get root path from [%s]\n", pkg);
        return 1;
      }
    }
  }
  else if (dll_mode)
  {
    for (const char* dll : args)
    {
      create_ni_select(dll);
    }
  }
  else
  {
    create_ni_under_dirs(args.data(), args.size());
  }

  return 0;
}
