/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <dlfcn.h>

#include <string>
#include <fstream>
#include <vector>

#include "utils.h"
#include "log.h"
#include "launcher.h"
#include "dotnet_launcher.h"

namespace tizen {
namespace runtime {
namespace dotnetcore {

CoreRuntime::CoreRuntime() :
  InitializeClr(nullptr),
  ExecuteAssembly(nullptr),
  Shutdown(nullptr),
  CreateDelegate(nullptr),
  coreclrLib(nullptr),
  hostHandle(nullptr),
  domainId(-1),
  PreparedFunction(nullptr),
  LaunchFunction(nullptr)
{
#define __XSTR(x) #x
#define __STR(x) __XSTR(x)

#ifdef DEVICE_API_DIR
  DeviceAPIDirectory = __STR(DEVICE_API_DIR);
#endif
#ifdef RUNTIME_DIR
  RuntimeDirectory = __STR(RUNTIME_DIR);
#endif
#ifdef NATIVE_LIB_DIR
  NativeLibDirectory = __STR(NATIVE_LIB_DIR);
#endif

#ifdef USE_MANAGED_LAUNCHER
#ifdef CORECLR_LAUNCHER_ASSEMBLY_PATH
  LauncherAssembly = __STR(CORECLR_LAUNCHER_ASSEMBLY_PATH);
#endif
#endif

#undef __STR
#undef __XSTR

  _DBG("Constructor called!!");
}

CoreRuntime::~CoreRuntime()
{
  Dispose();
}

int CoreRuntime::Initialize(bool standalone)
{

  if (standalone)
  {
    const char *_deviceapi_directory = getenv("DeviceAPIDirectory");
    const char *_runtime_directory = getenv("RuntimeDirectory");
    if (_deviceapi_directory != nullptr)
      DeviceAPIDirectory = _deviceapi_directory;
    if (_runtime_directory != nullptr)
      RuntimeDirectory = _runtime_directory;

#ifdef USE_MANAGED_LAUNCHER
    const char *_launcher_assembly = getenv("LauncherAssembly");
    if (_launcher_assembly != nullptr)
      LauncherAssembly = _launcher_assembly;
#endif
  }

  if (DeviceAPIDirectory.empty())
  {
    _ERR("Empty Device API Directory");
    return 1;
  }
  else
  {
    DeviceAPIDirectory = AbsolutePath(DeviceAPIDirectory);
  }

  if (RuntimeDirectory.empty())
  {
    _ERR("Empty Runtime Directory");
    return 1;
  }
  else
  {
    RuntimeDirectory = AbsolutePath(RuntimeDirectory);
  }

#ifdef USE_MANAGED_LAUNCHER
  if (LauncherAssembly.empty())
  {
    _ERR("Empty Launcher Assembly");
    return 1;
  }
  else
  {
    LauncherAssembly = AbsolutePath(LauncherAssembly);
  }
#endif

  std::string libcoreclr(ConcatPath(RuntimeDirectory, "libcoreclr.so"));

  _DBG("libcoreclr : %s", libcoreclr.c_str());

  coreclrLib = dlopen(libcoreclr.c_str(), RTLD_NOW | RTLD_LOCAL);
  if (coreclrLib == nullptr)
  {
    char *err = dlerror();
    _ERR("dlopen failed to open libcoreclr.so with error %s", err);
    return 1;
  }

#define CORELIB_RETURN_IF_NOSYM(type, variable, name) \
  do { variable = (type)dlsym(coreclrLib, name); \
    if (variable == nullptr) { \
      _ERR(name " is not found in the libcoreclr.so"); \
      return 1; \
    }} while(0)

  CORELIB_RETURN_IF_NOSYM(coreclr_initialize_ptr, InitializeClr, "coreclr_initialize");
  CORELIB_RETURN_IF_NOSYM(coreclr_execute_assembly_ptr, ExecuteAssembly, "coreclr_execute_assembly");
  CORELIB_RETURN_IF_NOSYM(coreclr_shutdown_ptr, Shutdown, "coreclr_shutdown");
  CORELIB_RETURN_IF_NOSYM(coreclr_create_delegate_ptr, CreateDelegate, "coreclr_create_delegate");

#undef CORELIB_RETURN_IF_NOSYM

  _DBG("libcoreclr dlopen and dlsym success");

  _DBG("this addr : %x", this);
  _DBG("coreclr_initialize : %x", InitializeClr);

  return 0;
}

bool CoreRuntime::InitializeCoreClr(const char* app_id,
                                    const char* assembly_probe_paths,
                                    const char* pinvoke_probe_paths,
                                    const char* tpa_list)
{
  const char *propertyKeys[] =
  {
    "TRUSTED_PLATFORM_ASSEMBLIES",
    "APP_PATHS",
    "APP_NI_PATHS",
    "NATIVE_DLL_SEARCH_DIRECTORIES",
    "AppDomainCompatSwitch"
  };

  const char *propertyValues[] =
  {
    tpa_list,
    assembly_probe_paths,
    assembly_probe_paths,
    pinvoke_probe_paths,
    "UseLatestBehaviorWhenTFMNotSpecified"
  };

  std::string selfPath = ReadSelfPath();

  int st = InitializeClr(
      selfPath.c_str(),
      app_id,
      sizeof(propertyKeys) / sizeof(propertyKeys[0]),
      propertyKeys,
      propertyValues,
      &hostHandle,
      &domainId);

  if (st < 0)
  {
    _ERR("initialize core clr fail! (0x%08x)", st);
    return false;
  }

  _DBG("Initialize core clr success");
  return true;
}

int CoreRuntime::RunManagedLauncher(const char* app_id, const char* app_base, const char* tpa_list)
{
  if (FileNotExist(LauncherAssembly))
  {
    _ERR("Launcher assembly is not exist in %s", LauncherAssembly.c_str());
    return 1;
  }

  if (!InitializeCoreClr(app_id, app_base, app_base, tpa_list))
  {
    _ERR("Failed to initialize coreclr");
    return 1;
  }

#ifdef USE_MANAGED_LAUNCHER
  void *preparedFunctionDelegate;
  int st = CreateDelegate(hostHandle, domainId,
      "Tizen.Runtime.Coreclr",
      "Tizen.Runtime.Coreclr.AssemblyManager",
      "Prepared", &preparedFunctionDelegate);
  if (st < 0)
  {
    _ERR("Create delegate for Launch prepared function is fail (0x%08x)", st);
    return 1;
  }
  PreparedFunction = reinterpret_cast<PreparedFunctionPtr>(preparedFunctionDelegate);

  if(PreparedFunction != nullptr)
  {
    PreparedFunction();
  }

  void *launchFunctionDelegate;
  st = CreateDelegate(hostHandle, domainId,
      "Tizen.Runtime.Coreclr",
      "Tizen.Runtime.Coreclr.AssemblyManager",
      "Launch", &launchFunctionDelegate);
  if (st < 0)
  {
    _ERR("Create delegate for Launch managed function is fail! (0x%08x)", st);
    return 1;
  }
  LaunchFunction = reinterpret_cast<LaunchFunctionPtr>(launchFunctionDelegate);
#endif
  return 0;
}

void CoreRuntime::Dispose()
{
  if (hostHandle != nullptr)
  {
    int st = Shutdown(hostHandle, domainId);
    if (st < 0)
    {
      _ERR("shutdown core clr fail! (0x%08x)", st);
    }
  }

  if (dlclose(coreclrLib) != 0)
  {
    _ERR("libcoreclr.so close failed");
  }
  coreclrLib = nullptr;

  _DBG("Dotnet runtime disposed");
}

int CoreRuntime::Launch(const char* app_id, const char* root, const char* path, int argc, char* argv[])
{
  if (path == nullptr)
  {
    _ERR("executable path is null");
    return 1;
  }

  if (FileNotExist(path))
  {
    _ERR("File not exist : %s", path);
    return 1;
  }

  std::vector<std::string> searchDirectories = {
    RuntimeDirectory, DeviceAPIDirectory
#ifdef USE_MANAGED_LAUNCHER
    , Basename(LauncherAssembly)
#endif
  };

  //std::string trusted_directories = JoinStrings(searchDirectories, ":");
  std::string tpa;
  AssembliesInDirectory(searchDirectories, tpa);

  std::string appRoot = root;
  std::string appBin = ConcatPath(appRoot, "bin");
  std::string appLib = ConcatPath(appRoot, "lib");
  std::string probePath = appBin + ":" + appLib + ":" + NativeLibDirectory;

#ifdef USE_MANAGED_LAUNCHER
  RunManagedLauncher(app_id, probePath.c_str(), tpa.c_str());

  bool success = false;
  if (LaunchFunction != nullptr)
  {
    std::string cpppath(path);

    if (IsManagedAssembly(cpppath) && !IsNativeImage(cpppath))
    {
      size_t extindex = cpppath.size() - 4;
      cpppath = cpppath.substr(0, extindex) + ".ni" + cpppath.substr(extindex, 4);
      if (!FileNotExist(cpppath))
      {
        path = cpppath.c_str();
      }
    }

    success = LaunchFunction(root, path, argc, argv);
    if (!success)
    {
      _ERR("Failed to launch Application %s", path);
    }
    return success ? 0 : 1;
  }
  else
  {
    _ERR("Failed to find launch function");
    return 1;
  }
#else
  int st = InitializeCoreClr(app_id, probePath.c_str(), probePath.c_str(), tpa.c_str());
  unsigned int ret = 0;
  st = ExecuteAssembly(hostHandle, domainId, argc, (const char**)argv, path, &ret);
  if (st < 0)
  {
    _ERR("Failed to Execute Assembly %s (0x%08x)", path, st);
  }
  return ret;
#endif
}

}  // namespace dotnetcore
}  // namespace runtime
}  // namespace tizen
