/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "launcher.h"
#include "log.h"

#include <launchpad.h>
#include <aul.h>

#include <Ecore.h>
#include <Elementary.h>
#include <bundle_internal.h>

#include <map>
#include <vector>
#include <functional>

#include <unistd.h>
#include <dlfcn.h>


namespace tizen {
namespace runtime {

struct FdHandler
{
  Ecore_Fd_Handler *handler;
  loader_receiver_cb receiver;
};

static int __argc;
static char **__argv;
static Evas_Object *__win;

class LaunchpadAdapterImpl : public LaunchpadAdapter
{
  public:
    LaunchpadAdapterImpl() : isLaunched(false) { }
    void LoaderMain(int argc, char* argv[]) override;

    std::map<int, FdHandler> Handlers;

  private:
    AppInfo appinfo;
    loader_lifecycle_callback_s callbacks;
    loader_adapter_s adapter;
    LauncherInterface* launcher;
    bool isLaunched;
    std::string launchPath;
};

LaunchpadAdapterImpl LaunchpadImpl;
LaunchpadAdapter& Launchpad = LaunchpadImpl;

#define WITH_SELF(data) \
  LaunchpadAdapterImpl* self = static_cast<LaunchpadAdapterImpl*>(data); \
  if (self == nullptr) \
  { \
    _ERR("No LaunchpadImplData"); \
  } else

static Eina_Bool Fd_Handler(void *data, Ecore_Fd_Handler* handler)
{
  WITH_SELF(data)
  {
    int fd = ecore_main_fd_handler_fd_get(handler);
    if (fd == -1)
    {
      _ERR("Failed to get the Ecore FD");
      exit(-1);
    }

    if (ecore_main_fd_handler_active_get(handler, ECORE_FD_READ))
    {
      if (self->Handlers.find(fd) != self->Handlers.end())
      {
        self->Handlers[fd].receiver(fd);
      }
    }
    else if (ecore_main_fd_handler_active_get(handler, ECORE_FD_ERROR))
    {
      _ERR("Ecore FD Handler Have Error");
      close(fd);
      exit(-1);
    }
  }

  return ECORE_CALLBACK_CANCEL;
}

static void Fd_Add(void *data, int fd, loader_receiver_cb receiver)
{
  Ecore_Fd_Handler* handler = ecore_main_fd_handler_add(fd,
      static_cast<Ecore_Fd_Handler_Flags>(ECORE_FD_READ | ECORE_FD_ERROR),
      Fd_Handler, data, nullptr, nullptr);
  if (handler == nullptr)
  {
    _ERR("Failed to add a FD handler to ecore main loop");
    close(fd);
    exit(-1);
  }
  WITH_SELF(data)
  {
    self->Handlers[fd] = {handler, receiver};
  }
}

static void Fd_Remove(void *data, int fd)
{
  WITH_SELF(data)
  {
    if (self->Handlers.find(fd) != self->Handlers.end())
    {
      Ecore_Fd_Handler* handler = self->Handlers[fd].handler;
      ecore_main_fd_handler_del(handler);
      self->Handlers.erase(fd);
    }
  }
}


static void PreloadLibsAndWindow(bundle *extra, int type, void *user_data)
{
	int elm_init_cnt = 0;
  const char **so_array;
  int len = 0;
  int i;
  void *handle = NULL;

  // Preload native libraries
	if (extra == NULL) {
		_DBG("No extra data");
		return;
	}

	so_array = bundle_get_str_array(extra, "preload", &len);

	if (!so_array)
		return;

	for (i = 0; i < len; i++) {
		handle = dlopen(so_array[i], RTLD_NOW);
		_DBG("preload %s# - handle : %x", so_array[i], handle);
	}

  // Precreate window
	elm_init_cnt = elm_init(__argc, __argv);
	_DBG("[candidate] elm init, returned: %d", elm_init_cnt);

  elm_config_accel_preference_set("hw");

  __win = elm_win_add(NULL, "package_name", ELM_WIN_BASIC);
  if (__win == NULL) {
    _DBG("[candidate] elm_win_add() failed");
    return;
  }

  elm_win_precreated_object_set(__win);
}

void LaunchpadAdapterImpl::LoaderMain(int argc, char* argv[])
{
  __argc = argc;
  __argv = argv;
  callbacks.create = [](bundle *extra, int type, void *user_data)
  {
    ecore_init();
    PreloadLibsAndWindow(extra, type, user_data);
    WITH_SELF(user_data)
    {
      if (self->OnCreate != nullptr)
        self->OnCreate();
    }
  };
  callbacks.launch = [](int argc, char** argv, const char* app_path,
      const char* appid, const char* pkgid,
      const char* pkg_type, void* user_data) -> int
  {
    WITH_SELF(user_data)
    {
      self->appinfo.root = std::string(aul_get_app_root_path());
      self->appinfo.path = app_path;
      self->appinfo.id = appid;
      self->appinfo.pkg = pkgid;
      self->appinfo.type = pkg_type;
      if (self->OnLaunch != nullptr)
        self->OnLaunch(self->appinfo, argc, argv);
    }

    return 0;
  };
  callbacks.terminate = [](int argc, char **argv, void* user_data) -> int
  {
    _DBG("Terminate!!");
    WITH_SELF(user_data)
    {
      if (self->OnTerminate != nullptr)
        self->OnTerminate(self->appinfo, argc, argv);
    }
    return 0;
  };

  adapter.loop_begin = [](void *data)
  {
    ecore_main_loop_begin();
  };

  adapter.loop_quit = [](void *data)
  {
    ecore_main_loop_quit();
  };
  adapter.add_fd = Fd_Add;
  adapter.remove_fd = Fd_Remove;

  _DBG("launchpad_loader_main is start");
  int r = launchpad_loader_main(argc, argv, &(this->callbacks), &(this->adapter), this);
  _DBG("launchpad_loader_main is finished with [%d]", r);
}

#undef WITH_SELF

}  // namespace runtime
}  // namespace tizen
